import numpy as np
import numba
import tqdm
import turbulence


def quiver_coordsum(coord, index, scale=3e-4, units='xy', extra=False,
                    mode='resid', **kw):
    from matplotlib import pyplot as p
    i = index
    if mode == 'resid':
        dx = coord['FPA_X'][:, i] - coord['EXPECTED_X'] - coord['XTURB'][:, i]
        dy = coord['FPA_Y'][:, i] - coord['EXPECTED_Y'] - coord['YTURB'][:, i]
    elif mode == 'total':
        dx = coord['FPA_X'][:, i] - coord['EXPECTED_X']
        dy = coord['FPA_Y'][:, i] - coord['EXPECTED_Y']
    elif mode == 'turbulence':
        dx = coord['XTURB'][:, i]
        dy = coord['YTURB'][:, i]
    else:
        raise ValueError('need to set mode to something understood.')
    q = p.quiver(coord['EXPECTED_X'], coord['EXPECTED_Y'], dx, dy,
                 scale=scale, units=units, **kw)
    p.gca().set_aspect('equal')
    m = ((coord['POSTYPE'] == 0) & np.isfinite(coord['FPA_X'][:, i]) &
         np.isfinite(coord['EXPECTED_X']))
    sigma = np.sqrt(np.nanmean(dx[m]**2 + dy[m]**2))
    p.text(120, 380, rf'$\sigma = {sigma*1000:3.1f}'+r'~\mathrm{\mu m}$')
    p.quiverkey(q, 0.12, 0.85, 10/1000, r'10 $\mathrm{\mu m}$')
    if extra:
        p.xlabel('X (mm)')
        p.ylabel('Y (mm)')
    return q, sigma


def make_expected_positions(coordsum, empirical=False, fac=1):
    ctrldisabled = (coordsum['FLAGS_COR'] & 2**16) != 0
    validdata = np.isfinite(coordsum['FPA_X'])
    xmean = np.nanmedian(coordsum['FPA_X'], axis=1)
    ymean = np.nanmedian(coordsum['FPA_Y'], axis=1)
    # xstd = mad_std(coordsum['FPA_X'], axis=1, ignore_nan=True)
    # ystd = mad_std(coordsum['FPA_Y'], axis=1, ignore_nan=True)
    xstd = np.nanstd(coordsum['FPA_X'], axis=1)
    ystd = np.nanstd(coordsum['FPA_Y'], axis=1)
    if empirical:
        stationary = (xstd > 0) & (xstd < fac*1e-2) & (ystd < fac*1e-2)
    else:
        stationary = np.all(ctrldisabled | ~validdata, axis=1)
    stationary = stationary & np.any(validdata, axis=1)
    fiducial = np.any(((coordsum['FLAGS_COR'] & 2**30) != 0) &
                      validdata, axis=1) & stationary
    # fiducial = fiducial & (xstd < 0.03) & (ystd < 0.03)
    postype = 2**0*stationary + 2**1*fiducial
    return xmean, ymean, postype


def data_file_list(outfile):
    # 20240124, exposures 221162-221229, 221262-221292
    # F0000 through F0011
    fd = open(outfile, 'w')
    exps = list(range(221162, 221230)) + list(range(221262, 221293))
    for exp in exps:
        for i in range(12):
            fd.write(f'{exp}, {i:04d}\n')
    fd.close()
# compute all positions via output of this file and this bash invocation on an
# interactive node
# cat flist.csv | parallel -j 60 --colsep ',' desi_fvc_proc -i $DESI_SPECTRO_DATA/20240124/00{1}/fvc-00{1}.fits.fz --extname F{2} --expected-positions $DESI_SPECTRO_DATA/20240124/00{1}/coordinates-00{1}.fits --fixed-rotation -o fvc-00{1}-{2}.csv
# takes about 5 minutes.
# did not actually use this; desimeter turbulence estimates have some weird
# bimodality that I did not want to learn how to understand.


def correlation_function(coord, spacing, each=False):
    """Compute correlation function of turbulence measurements given edges.

    We need to generate the parallel and perpendicular correlation functions.
    Loop over all exposures.
    For each exposure, match all points to one another at some distance.
    Or just do the full all pairs i > j?  Sounds fine.  Probably numba.
    Could parallelize by exposure.
    Fill out the sums.

    Parameters
    ----------
    coord: coordinates summary file from gather_nights
    spacing: bin spacing

    Returns
    -------
    vpar, vperp, npair
    """

    xmean, ymean, postype = make_expected_positions(coord)
    dx = coord['FPA_X'] - xmean[:, None]
    dy = coord['FPA_Y'] - ymean[:, None]
    m = np.isfinite(xmean)
    dx = dx[m, :].astype('<f4')
    dy = dy[m, :].astype('<f4')
    x = coord['FPA_X'][m, :].astype('<f4')
    y = coord['FPA_Y'][m, :].astype('<f4')
    nexp = dx.shape[1]

    # units: mm on focal plane
    # maxbin = 450 / spacing
    maxbin = int(450 * 2 / spacing) + 2

    vpar = np.zeros((maxbin + 1, nexp), dtype='f4')
    vperp = np.zeros((maxbin + 1, nexp), dtype='f4')
    npair = np.zeros((maxbin + 1, nexp), dtype='i4')

    for i in tqdm.tqdm(range(nexp)):
        vpar0, vperp0, npair0 = compute_correlation_function(
            x[:, i], y[:, i], dx[:, i], dy[:, i], spacing, maxbin)
        vpar[:, i] = vpar0
        vperp[:, i] = vperp0
        npair[:, i] = npair0
    bins = spacing / 2 + np.arange(maxbin + 1)

    if not each:
        vpar = np.sum(vpar, axis=1)
        vperp = np.sum(vperp, axis=1)
        npair = np.sum(npair, axis=1)

    return (vpar / (npair + (npair == 0)),
            vperp / (npair + (npair == 0)),
            npair,
            bins)


@numba.jit(nopython=True, cache=False)
def binfunction(sep, spacing):
    return int(sep / spacing)


@numba.jit(nopython=True, cache=False)
def compute_correlation_function(x, y, dx, dy, spacing, maxbin):
    vpar = np.zeros(maxbin + 1, dtype='f4')
    vperp = np.zeros(maxbin + 1, dtype='f4')
    npair = np.zeros(maxbin + 1, dtype='i4')
    for i in range(len(x)):
        for j in range(i + 1, len(x)):
            dxi = dx[i]
            dyi = dy[i]
            dxj = dx[j]
            dyj = dy[j]
            vecx = x[i] - x[j]
            vecy = y[i] - y[j]
            sep = (vecx**2 + vecy**2)**0.5
            vecxn = vecx / sep
            vecyn = vecy / sep
            index = binfunction(sep, spacing)
            vpari = vecxn * dxi + vecyn * dyi
            vparj = vecxn * dxj + vecyn * dyj
            vperpi = -vecyn * dxi + vecxn * dyi
            vperpj = -vecyn * dxj + vecxn * dyj
            npair[index] += 1
            vpar[index] += vpari * vparj
            vperp[index] += vperpi * vperpj
    return vpar, vperp, npair


# this works, gets some functions for vpar, vperp
# they don't quite satisfy vpar = d/dr (r * vperp)
# not clear if I have a bug or if this is something interesting.
# how do things like the integral constraint come in here?
# we have other interesting things like "missing power" on large
# scales due to the initial PM fit.
# Maybe we could do something like go back to the raw x/y but it's
# not clear that that's ultimately beneficial.
# match is quite good on small scales (right on at r = 0; 2% at r=50 mm,
# 6% at r=100 mm) but degrades going larger (factor of 2 at r=200).
# still, we should be able to infer the correlation function for the
# scalar field.


def plot_correlation_function(vpara, vperpa, npaira, bins):
    from matplotlib import pyplot as p
    p.figure()
    p.gcf().set_size_inches(4, 3.5, forward=True)
    qpar16, qpar50, qpar84 = np.quantile(vpara, axis=1, q=[0.16, 0.5, 0.84])
    qperp16, qperp50, qperp84 = np.quantile(vperpa, axis=1, q=[0.16, 0.5, 0.84])
    npair = np.sum(npaira, axis=1)
    vpar = np.sum(vpara * npaira, axis=1) / (npair + (npair == 0))
    vperp = np.sum(vperpa * npaira, axis=1) / (npair + (npair == 0))
    p.plot(bins, vpar * 10**6, label=r'$\Psi_\parallel$', color='C0')
    p.plot(bins, vperp * 10**6, label=r'$\Psi_\perp$', color='C1')
    p.fill_between(bins, qpar16 * 10**6, qpar84 * 10**6, alpha=0.3,
                   color=f'C0', edgecolor='none')
    p.fill_between(bins, qperp16 * 10**6, qperp84 * 10**6, alpha=0.3,
                   color=f'C1', edgecolor='none')

    p.xlim(6, 820)
    p.ylim(-2e1, 7e1)
    coeff = np.polyfit(bins, vperp, deg=10, w=npair)
    p.plot(bins, np.gradient(np.polyval(coeff, bins) * bins * 10**6, bins),
           linestyle='--', color='C0',
           label=r'$\frac{d}{dr}(r \Psi_\perp) \approx \Psi_\parallel$')
    p.legend()
    p.xlabel('separation (mm)')
    p.ylabel('correlation function $\Psi$ ($\mu$m$^2$)')
    p.subplots_adjust(left=0.2, bottom=0.15)
    p.axhline(0, linestyle=':', color='gray')
    p.savefig('correlation.pdf')


def turbulence_example(coord, index=0):
    from matplotlib import pyplot as p
    p.figure()
    p.gcf().set_size_inches(4, 4, forward=True)
    quiver_coordsum(coord, index, mode='total', extra=True)
    p.subplots_adjust(left=0.18, right=0.98, bottom=0.13, top=0.9)
    p.savefig('turbexample.pdf')


def turbulence_gallery(coord, index=[1, 2, 3]):
    from matplotlib import pyplot as p
    p.figure()
    p.gcf().set_size_inches(7.4, 7.5, forward=True)
    p.subplots_adjust(wspace=0, hspace=0, left=0.1, right=0.98, bottom=0.08,
                      top=0.95)
    for count, i in enumerate(index):
        p.subplot(3, len(index), count + 1)
        quiver_coordsum(coord, i, mode='total', extra=False)
        if count != 0:
            p.gca().yaxis.set_ticklabels([])
        else:
            p.ylabel('Y (m)')
        p.gca().xaxis.set_ticklabels([])
        p.subplot(3, len(index), 3 + count + 1)
        quiver_coordsum(coord, i, mode='turbulence', extra=False)
        if count != 0:
            p.gca().yaxis.set_ticklabels([])
        else:
            p.ylabel('Y (m)')
        p.gca().xaxis.set_ticklabels([])
        p.subplot(3, len(index), 6 + count + 1)
        quiver_coordsum(coord, i, mode='resid', extra=False)
        if count != 0:
            p.gca().yaxis.set_ticklabels([])
        else:
            p.ylabel('Y (m)')
        p.xlabel('X (mm)')
    p.savefig('turbulencegallery.pdf')


def posperformance(coordstats):
    from matplotlib import pyplot as p
    import util_efs
    p.figure()
    p.gcf().set_size_inches(4, 3.5, forward=True)
    for i, field in enumerate(
            ['rms_total', 'rms_turbulence', 'rms_positioning']):
        dat = coordstats[field]*1000
        val = np.nanmedian(dat)
        fieldname = field.replace('_', ' ')
        label = rf'{fieldname}, $\sigma = {val:3.1f}'+r' \mathrm{~\mu m}$'
        scat = util_efs.Scatter(
            coordstats['expid'] / 10000, dat, yrange=[0, 15],
            xrange=[11, 23])
        p.fill_between(scat.xpts, scat.q84, scat.q16, label=label, alpha=0.3,
                       color=f'C{i}')
        p.axhline(val, color=f'C{i}', linestyle='--')
    p.legend()
    p.xlabel('Exposure ID / 10000')
    p.subplots_adjust(bottom=0.15, left=0.15)
    p.ylabel(r'RMS positioning error ($\mathrm{\mu m}$)')
    p.ylim(2.1, 17.5)
    p.savefig('positioningstats.pdf')


def fill_in_turbulence_estimates(
        coordsum, method='independent', empirical=False, firstn=np.inf,
        nstationary=None):
    xmean, ymean, postype = make_expected_positions(
        coordsum, empirical=empirical)
    coordsum['EXPECTED_X'] = xmean
    coordsum['EXPECTED_Y'] = ymean
    coordsum['POSTYPE'] = postype
    n = int(min([firstn, coordsum['EXPID'].shape[1]]))

    if nstationary is not None:
        xmean, ymean, postype = make_expected_positions(
            coordsum, empirical=True)
        mcoulduse = postype != 0
        np.random.seed(1)
        iuse = np.random.choice(np.flatnonzero(mcoulduse),
                                size=nstationary,
                                replace=False)
        m = np.zeros_like(postype, dtype='bool')
        m[iuse] = True
        postype[~m] = 0
        # don't use not-subselected positioners for turbulence correction.
        if not empirical:
            raise ValueError('do not set nstationary and not empirical.')

    coordsum['EXPECTED_X'] = xmean
    coordsum['EXPECTED_Y'] = ymean
    coordsum['POSTYPE'] = postype

    for i in tqdm.tqdm(range(n)):
        valid = np.isfinite(coordsum['FPA_X'][:, i])
        muse = valid & (postype != 0)
        minterp = valid
        if np.sum(valid) == 0:
            continue
        xturb, yturb = turbulence.correct_using_stationary_test(
            coordsum['FPA_X'][muse, i],
            coordsum['FPA_Y'][muse, i],
            xmean[muse], ymean[muse],
            coordsum['FPA_X'][minterp, i],
            coordsum['FPA_Y'][minterp, i],
            method=method)
        coordsum['XTURB'][minterp, i] = xturb
        coordsum['YTURB'][minterp, i] = yturb


def stuck_positioner_performance(coord):
    c = coord.copy()
    from astropy.stats import sigma_clipped_stats
    modes = ['none', 'simple', 'independent', 'gradwavefront']
    nposoptiondict = {30: dict(nstationary=30, empirical=True),
                      100: dict(nstationary=100, empirical=True),
                      300: dict(nstationary=300, empirical=True),
                      'DESI': dict()}
    out = dict()
    for npos, options in nposoptiondict.items():
        for mode in modes:
            fill_in_turbulence_estimates(
                coord, method=mode, **options)
            c = coord
            m = (c['POSTYPE'] == 0) & np.isfinite(c['FPA_X'][:, 0])
            dx = c['FPA_X'][m, :] - c['EXPECTED_X'][m, None] - c['XTURB'][m, :]
            dy = c['FPA_Y'][m, :] - c['EXPECTED_Y'][m, None] - c['YTURB'][m, :]
            print('nstationary: ',
                  np.sum((c['POSTYPE'] != 0) & np.isfinite(c['FPA_X'][:, 0])))
            mn, med, sigma = sigma_clipped_stats(dx**2 + dy**2, axis=0)
            val = np.median(np.sqrt(mn))
            out[(npos, mode)] = val
    return out


def stuck_positioner_performance_plot(performancedict):
    from matplotlib import pyplot as p
    from copy import deepcopy
    p.figure()
    p.gcf().set_size_inches(4, 3.5, forward=True)
    nposall = list(set([k[0] for k in performancedict]))
    modeall = ['none', 'simple', 'independent', 'gradwavefront']
    for mode in modeall:
        pts = deepcopy(nposall)
        pts[pts == 'DESI'] = 793
        vals = []
        for npos in nposall:
            vals.append(performancedict[(npos, mode)])
        vals = np.array(vals)*1000
        s = np.argsort(pts)
        pts = np.array(pts)[s]
        vals = vals[s]
        p.plot(pts, vals, '-+', label=mode)
    p.xlabel('Number of turbulence-probe positioners')
    p.ylabel(r'Positioning error $\mathrm{\mu m}$')
    p.gca().xaxis.set_ticks(pts)
    xticklabels = [x.get_text() for x in p.gca().xaxis.get_ticklabels()]
    xticklabels[-1] = xticklabels[-1] + ' (DESI)'
    p.gca().xaxis.set_ticklabels(xticklabels)
    p.subplots_adjust(bottom=0.15, top=0.95)
    p.legend()
    p.savefig('nstuckperformance.pdf')


def doall():
    from astropy.io import fits
    from astropy.table import Table
    coord = fits.getdata('coord-summary-20240124-pm.fits')
    fill_in_turbulence_estimates(coord, method='independent')
    coordstats = Table.read('coordinates-stats.ecsv')
    turbulence_example(coord)
    turbulence_gallery(coord)
    posperformance(coordstats)
    vpar, vperp, npair, bins = correlation_function(coord, 1, each=True)
    plot_correlation_function(vpar, vperp, npair, bins)
    perfdict = stuck_positioner_performance(coord)
    stuck_positioner_performance_plot(perfdict)
