import numpy as np
import astropy.io.ascii
from scipy import optimize
import scipy.linalg
from desimeter.match_positioners import match


def match2d_nearest(x1, y1, x2, y2, rad):
    """Find nearest match between x1, y1 and x2, y2 within radius rad."""
    from scipy.spatial import cKDTree
    xx1 = np.stack([x1, y1], axis=1)
    xx2 = np.stack([x2, y2], axis=1)
    tree1 = cKDTree(xx1)
    dd, ii = tree1.query(xx2, distance_upper_bound=rad)
    m = np.isfinite(dd)
    return ii[m], np.flatnonzero(m), dd[m]


def make_covar_gradwavefront(data1, param, data2=None, rq=False,
                             nonoise=False):
    ll = param[2]
    if data2 is None:
        data2 = data1
    xdist = (data1['x'][None, :] - data2['x'][:, None])/ll
    ydist = (data1['y'][None, :] - data2['y'][:, None])/ll
    nn1 = len(data1['x'])
    nn2 = len(data2['x'])
    covar = np.empty((nn2*2, nn1*2), dtype='f4')
    if not rq:
        sigma, aa, ll = param
        gauss = aa**2*np.exp(-(xdist**2+ydist**2)/2) / ll**2
        covar[:nn2, :nn1] = (1-xdist*xdist)*gauss
        covar[nn2:, :nn1] = -xdist*ydist*gauss
        covar[:nn2, nn1:] = -xdist*ydist*gauss
        covar[nn2:, nn1:] = (1-ydist*ydist)*gauss
    else:
        sigma, aa, ll, alpha = param
        xdist /= np.sqrt(np.abs(alpha))
        ydist /= np.sqrt(np.abs(alpha))
        fac = 1 + (xdist**2 + ydist**2)/2
        prefac = fac**(-alpha - 2) * aa**2 / ll**2
        covar[:nn2, :nn1] = prefac * ((alpha + 1)*xdist**2 - fac)
        covar[nn2:, nn1:] = prefac * ((alpha + 1)*ydist**2 - fac)
        covar[:nn2, nn1:] = prefac * (alpha + 1) * xdist * ydist
        covar[nn2:, :nn1] = prefac * (alpha + 1) * xdist * ydist
    if not nonoise:
        if nn1 != nn2:
            raise ValueError
        covar += np.diag(np.full(2*nn1, sigma**2))
    return covar


def loss_gradwavefront(data, param, rq=False):
    covar = make_covar_gradwavefront(data, param, rq=rq)
    chol, low = scipy.linalg.cho_factor(covar, check_finite=False,
                                        overwrite_a=True)
    dvec = np.concatenate([data['dx'], data['dy']])
    cinvd = scipy.linalg.cho_solve((chol, low), dvec, check_finite=False)
    chi2d = np.dot(dvec, cinvd)
    logdetcovar = 2*np.sum(np.log(np.diag(chol)))
    chi2 = chi2d + logdetcovar
    return 0.5*chi2


def make_covar_independent(data, param, data2=None, rq=False, nonoise=False):
    # might be good to have more large-scale coherence
    # "rational quadratic" seems like the next obvious choice.
    if not rq:
        sigma, aa, ll = param
    else:
        sigma, aa, ll, alpha = param
        alpha = np.clip(alpha, 0.1, 1000)
    if data2 is None:
        data2 = data
    dist2 = ((data['x'][None, :] - data2['x'][:, None])**2 +
             (data['y'][None, :] - data2['y'][:, None])**2)/ll**2
    if not rq:
        covar = aa**2*np.exp(-dist2/2)
    else:
        covar = aa**2*(1+dist2/(2*alpha))**(-alpha)
    if not nonoise:
        covar += np.eye(len(data))*sigma**2
    return covar


def loss_independent(data, param, rq=False):
    covar = make_covar_independent(data, param, rq=rq)
    # chol = np.linalg.cholesky(covar)
    chol, low = scipy.linalg.cho_factor(covar, check_finite=False,
                                        overwrite_a=True)
    cinvx = scipy.linalg.cho_solve((chol, low), data['dx'], check_finite=False)
    cinvy = scipy.linalg.cho_solve((chol, low), data['dy'], check_finite=False)
    chix2 = np.dot(data['dx'], cinvx)
    chiy2 = np.dot(data['dy'], cinvy)

    # factor of 2 for cholesky, and another factor of 2 for x & y
    logdetcovar = 2*2*np.sum(np.log(np.diag(chol)))
    chi2 = chix2+chiy2+logdetcovar
    return 0.5*chi2


def make_data(expect, measure):
    mex, mme = match(expect['LOCATION'], measure['LOCATION'])
    data = np.zeros(len(mex), dtype=[
        ('location', 'i4'), ('x', 'f8'), ('y', 'f8'),
        ('dx', 'f8'), ('dy', 'f8')])
    data['location'] = expect['LOCATION'][mex]
    data['x'] = expect['X_FP'][mex]
    data['y'] = expect['Y_FP'][mex]
    data['dx'] = measure['X_FP'][mme] - data['x']
    data['dy'] = measure['Y_FP'][mme] - data['y']
    return data


def solve_covar(data, lossfun, covarfun, rq=False, nuse=500, guess=None,
                **kw):
    if nuse > 0:
        rr2 = data['x']**2+data['y']**2
        s = np.argsort(rr2)
        datause = data[s[:nuse]]
    else:
        datause = data

    def loss(param):
        res = lossfun(datause, param, rq=rq)
        print(param, res)
        return res

    # sigma, A, l
    if guess is None:
        guess = [0.01, 0.01, 10]
        if rq:
            guess = guess + [2]
    res = optimize.minimize(loss, guess, **kw)
    covar = covarfun(data, res.x, rq=rq)
    return covar, res


def solve_files(expectedfn, measuredfn, mode='independent', **kw):
    expect = astropy.io.ascii.read(expectedfn)
    measure = astropy.io.ascii.read(measuredfn)
    data = make_data(expect, measure)
    if mode == 'independent':
        xturb, yturb, res = solve_independent(data, **kw)
    elif mode == 'gradwavefront':
        xturb, yturb, res = solve_gradwavefront(data, **kw)
    else:
        raise ValueError('unknown mode')
    out = measure.copy()
    md, mo = match(data['location'], out['LOCATION'])
    newnames = ['XNEW', 'YNEW', 'DXTURB', 'DYTURB']
    for name in newnames:
        out[name] = np.nan
    out['XNEW'][mo] = out['X_FP'][mo]-xturb[md]
    out['YNEW'][mo] = out['Y_FP'][mo]-yturb[md]
    out['DXTURB'][mo] = xturb[md]
    out['DYTURB'][mo] = yturb[md]
    return out, res


def correct(x, y, x0, y0, dx=None, dy=None):
    data = np.zeros(len(x), dtype=[
        ('x', 'f8'), ('y', 'f8'),
        ('dx', 'f8'), ('dy', 'f8')])
    data['x'] = x
    data['y'] = y
    data['dx'] = x-x0
    data['dy'] = y-y0
    xturb, yturb, _ = solve_independent(data, nuse=500, excludeself=True)
    return x-xturb, y-yturb


def correct_using_stationary_test(
        xs, ys, x0s, y0s, xc, yc,
        scale_covar=1,
        return_good=False, method='none'):
    """Remove correlated turbulent signals from measured locations of fibers,
    using subset of stationary fibers.

    See `correct` for more documentation.  This function extends `correct`
    to use a subset of the fibers to correct all the fibers.

    Args:
        xs : array_like(n), measured x position of stationary fibers
        ys : array_like(n), measured y position of stationary fibers
        x0s : array_like(n), expected x position of stationary fibers
        y0s : array_like(n), expected y position of stationary fibers
        xc : array_like(n), measured x position of fibers to be corrected
        yc : array_like(n), measured y position of fibers to be corrected
        scale_covar : bool, amount to scale covariance by
            (pix / micron for FVC instead of FP, e.g.)

    Returns:
        xturb, yturb: derived turbulent offsets.  Turbulence corrected
        positions are given by xc-xturb, yc-yturb.
    """
    data = np.zeros(len(xs), dtype=[
        ('x', 'f8'), ('y', 'f8'),
        ('dx', 'f8'), ('dy', 'f8')])
    data['x'] = xs
    data['y'] = ys
    data['dx'] = xs-x0s
    data['dy'] = ys-y0s
    mask = (np.isfinite(data['x']) & np.isfinite(data['y']) &
            np.isfinite(data['dx']) & np.isfinite(data['dy']))
    data = data[mask]
    if method == 'none':
        xturb = xc * 0
        yturb = yc * 0
    elif method == 'simple':
        # just get the closest stationary measurement
        ms, mc, dsc = match2d_nearest(xs, ys, xc, yc, 200)
        xturb = xc * 0
        yturb = yc * 0
        xturb[mc] = data['dx'][ms]
        yturb[mc] = data['dy'][ms]
    elif method == 'independent':
        xturb, yturb, _ = solve_independent(
            data, excludeself=False,
            predict_at=(xc, yc),
            fix_covar=True, scale_covar=scale_covar)
    elif method == 'gradwavefront':
        xturb, yturb, _ = solve_gradwavefront(
            data, excludeself=False, predict_at=(xc, yc),
            fix_covar=True, scale_covar=scale_covar)
    out = (xturb, yturb)
    if return_good:
        out = out + (mask,)
    return out


def solve_independent(data, excludeself=False, predict_at=None,
                      fix_covar=False, scale_covar=1, **kw):
    """Find turbulent contributions to measured fiber positions.

    Assumes that the covariance matrix should be the same for x & y,
    but that otherwise the x & y directions are independent.

    Args:
        data : ndarray containing measured positions and residuals
            from expected locations
        excludeself : bool
            do not use this fiber when computing the turbulence
            affecting this fiber.
        predict_at : tuple(array_like, array_like)
            x & y coordinates of locations at which to predict turbulence
        **kw : additional keywords passed to solve_covar

    Returns:
        xturb : turbulent contributions in x direction
        yturb : turbulent contributions in y direction
        res : output from scipy.optimize.minimize describing best fit
            covariance matrix
    """
    if not fix_covar:
        covar, res = solve_covar(data, lossfun=loss_independent,
                                 covarfun=make_covar_independent, **kw)
    else:
        from types import SimpleNamespace
        res = SimpleNamespace()
        res.x = [5e-3*scale_covar, 5e-3*scale_covar, 50*scale_covar]
        if kw.get('rq', False):
            res.x = res.x + [2]
        covar = make_covar_independent(data, res.x, **kw)

    if predict_at is not None and excludeself:
        raise ValueError('predict_at does not make sense in combination with '
                         'excludeself')

    if not excludeself:
        if predict_at:
            # K(X*, X)(K(X, X) + C_n)^-1 y
            # Rasmussen & Williams algorithm 2.1
            chol, low = scipy.linalg.cho_factor(covar, check_finite=False,
                                                overwrite_a=True)
            covarpred = make_covar_independent(
                data, res.x, data2=dict(x=predict_at[0], y=predict_at[1]),
                nonoise=True, **kw)
            alphax = scipy.linalg.cho_solve((chol, low), data['dx'])
            alphay = scipy.linalg.cho_solve((chol, low), data['dy'])
            xturb = np.dot(covarpred, alphax)
            yturb = np.dot(covarpred, alphay)
        else:
            # covar includes measurement uncertainties; separate that part out.
            cninv = np.eye(len(data))*res.x[0]**(-2)
            covar -= np.eye(len(data))*res.x[0]**2
            cpcninv = np.dot(covar, cninv)
            aa = cpcninv+np.eye(len(data))
            xturb = np.linalg.solve(aa, np.dot(cpcninv, data['dx']))
            yturb = np.linalg.solve(aa, np.dot(cpcninv, data['dy']))
    else:
        # Rasmussen & Williams 5.12
        kinv = np.linalg.inv(covar)
        xturb = data['dx']-kinv.dot(data['dx'])/np.diag(kinv)
        yturb = data['dy']-kinv.dot(data['dy'])/np.diag(kinv)
    return xturb, yturb, res


def solve_gradwavefront(data, excludeself=False, predict_at=None,
                        fix_covar=False, scale_covar=1, **kw):
    """Find turbulent contributions to measured fiber positions.

    Assumes that the turbulent contributions can be modeled as the
    gradient of a wavefront error.  i.e., they are curl free.

    Args:
        data : ndarray containing measured positions and residuals
            from expected locations
        excludeself : bool
            do not use this fiber when computing the turbulence
            affecting this fiber.
        **kw : additional keywords passed to solve_covar

    Returns:
        xturb : turbulent contributions in x direction
        yturb : turbulent contributions in y direction
        res : output from scipy.optimize.minimize describing best fit
            covariance matrix
    """

    if predict_at is not None and excludeself:
        raise ValueError('predict_at does not make sense in combination with '
                         'excludeself')

    if not fix_covar:
        covar, res = solve_covar(data, lossfun=loss_gradwavefront,
                                 covarfun=make_covar_gradwavefront, **kw)
    else:
        from types import SimpleNamespace
        res = SimpleNamespace()
        res.x = [2e-3*scale_covar, 1*scale_covar, 100*scale_covar]
        covar = make_covar_gradwavefront(data, res.x, **kw)

    dvec = np.concatenate([data['dx'], data['dy']])
    if not excludeself:
        if predict_at:
            # K(X*, X)(K(X, X) + C_n)^-1 y
            # Rasmussen & Williams algorithm 2.1
            chol, low = scipy.linalg.cho_factor(covar, check_finite=False,
                                                overwrite_a=True)
            covarpred = make_covar_gradwavefront(
                data, res.x, data2=dict(x=predict_at[0], y=predict_at[1]),
                nonoise=True, **kw)
            alpha = scipy.linalg.cho_solve((chol, low), dvec)
            turb = np.dot(covarpred, alpha)
            xturb, yturb = turb[:len(predict_at[0])], turb[len(predict_at[0]):]
        else:
            # remove measurement noise contribution to covar
            cninv = np.eye(len(dvec))*res.x[0]**(-2)
            covar -= np.eye(len(dvec))*res.x[0]**2
            cpcninv = np.dot(covar, cninv)
            aa = cpcninv+np.eye(len(dvec))
            turb = np.linalg.solve(aa, np.dot(cpcninv, dvec))
            xturb, yturb = turb[:len(data)], turb[len(data):]
    else:
        # Rasmussen & Williams 5.12
        kinv = np.linalg.inv(covar)
        turb = dvec - kinv.dot(dvec)/np.diag(kinv)
        xturb, yturb = turb[:len(data)], turb[len(data):]
    return xturb, yturb, res


def empirical_covariance(data, bins=10, edges=None):
    dist = np.sqrt((data['x'][None, :] - data['x'][:, None])**2 +
                   (data['y'][None, :] - data['y'][:, None])**2).ravel()
    # meandx = np.mean(data['dx'])
    # meandy = np.mean(data['dy'])
    meandx = 0
    meandy = 0
    dx1dx2 = ((data['dx'][None, :] - meandx) *
              (data['dx'][:, None] - meandx)).ravel()
    dy1dy2 = ((data['dy'][None, :] - meandy) *
              (data['dy'][:, None] - meandy)).ravel()
    if edges is not None:
        bins = len(edges)-1
    else:
        edges = np.linspace(np.min(dist), np.max(dist)+0.1, bins+1)
    ind = np.searchsorted(edges, dist, side='right')-1
    m = ind >= 0
    ncount = np.bincount(ind[m], minlength=bins)
    ncount = ncount + (ncount == 0)
    dx1dx2 = np.bincount(ind[m], weights=dx1dx2[m], minlength=bins)
    dy1dy2 = np.bincount(ind[m], weights=dy1dy2[m], minlength=bins)
    return (edges[:-1]+edges[1:])/2, dx1dx2/ncount, dy1dy2/ncount


def turbulence_gallery(fn, expectfn):
    from matplotlib import pyplot as p
    p.clf()
    expect = astropy.io.ascii.read(expectfn)
    p.subplots_adjust(hspace=0., wspace=0.)
    for i, fn0 in enumerate(fn):
        measure = astropy.io.ascii.read(fn0)
        data = make_data(expect, measure)
        covar, res = solve_covar(data, lossfun=loss_independent,
                                 covarfun=make_covar_independent)
        print(res.x)
        uu, ss, _ = np.linalg.svd(covar)
        xvec = np.dot(uu, np.random.randn(len(ss))*np.sqrt(ss))
        yvec = np.dot(uu, np.random.randn(len(ss))*np.sqrt(ss))
        for k, (dx, dy) in enumerate([
                [data['dx'], data['dy']], [xvec, yvec]]):
            p.subplot(2, 3, 2*i+k+1)
            p.quiver(data['x'], data['y'], dx, dy,
                     units='x', scale=0.001)
            p.gca().xaxis.set_ticklabels('')
            p.gca().yaxis.set_ticklabels('')
            p.gca().xaxis.set_ticks([])
            p.gca().yaxis.set_ticks([])
            p.gca().set_aspect('equal')
